/**************************************
Copyright Elisabeth Tillier and Andrew Smith, University Health Network, 2003

 * PROGRAM TO CONVERT A 20 SYMBOL
 * CODED RNA SEQUENCE/STRUCTURE
 * BACK INTO THE 4 BASE RNA CODE.
 **************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

// The size of the input and output buffers.
#define BUFFER_SIZE 200000       
#define MAX_SEQ_LEN 200000

// Number of symbols in the original
// sequence (including structure).
#define CODE_COUNT 25           

// The input buffer for sequences.
char buffer[BUFFER_SIZE];

char *invert[26] = {
  /* A */  "A-",
  /* B */  "",
  /* C */  "AA",
  /* D */  "T-",
  /* E */  "AG",
  /* F */  "GC",
  /* G */  "AT",
  /* H */  "CA",
  /* I */  "CC",
  /* J */  "",
  /* K */  "CT",
  /* L */  "CG",
  /* M */  "GA",
  /* N */  "G-",
  /* O */  "",
  /* P */  "GG",
  /* Q */  "AC",
  /* R */  "C-",
  /* S */  "GT",
  /* T */  "TA",
  /* U */  "",
  /* V */  "TT",
  /* W */  "TC",
  /* X */  "XX",
  /* Y */  "TG",
  /* Z */  ""
};

/**********************************************
 *  CONVERT ANY DNA (T) TO RNA (U) AND MAKE SURE ALL
 *  ARE UPPER CASE
 **********************************************/
void upper_case(char *s) {
  int i = (int)strlen(s) - 1;
  for (; i >= 0; i--) s[i] = toupper(s[i]);
}

/********************************************************
 * READ A LINE OF TEXT FROM A FILE INTO A BUFFER
 ********************************************************/
int readLine(FILE *fp, char *buff, int buff_size) {
  int i;
  for (i = 0; i < buff_size && 
	 (buff[i] = (char)fgetc(fp)) != (char)EOF && 
	 (buff[i] != '\n') && (buff[i] != '\r'); i++);
  if (buff[i] == (char)EOF) return -1;
  buff[i] = '\0';
  return i;
}

/***********************************
 * CONVERT THE SEQUENCES INTO THE 20
 * SYMBOL CODE
 ***********************************/
void convert(char *original, FILE *fasta_out, FILE *phylip_out) {
  int length, i = 0;
  
  // Convert to upper case, and get rid of the 'U's
  upper_case(original);
  
  // Get the length of the original string.
  length = (int)strlen(original);
  
  // Convert each position...
  for (i = 0; i < length; i++) {
    fprintf(fasta_out, "%s", invert[(int)(original[i] - 'A')]);
    fprintf(phylip_out, "%s", invert[(int)(original[i] - 'A')]);
  }
  fprintf(fasta_out, "\n");
  fprintf(phylip_out, "\n");
}

/***********************************
 * READ IN THE RNA SEQUENCES FROM A
 * FASTA FORMAT FILE.
 ***********************************/
void convert_sequences(char *sequence_file, char *fasta_outfile, char *phylip_outfile) {
  FILE *in, *fasta_out, *phylip_out; 
  int sequence_count = 0, sequence_length = 0, flag = 0;
  char *input_sequence;

  // Open the input and output files.
  if (!(in = fopen(sequence_file, "r"))) {
    fprintf(stderr,"ERROR: bad sequence file name:\t%s\n", sequence_file);
    exit(EXIT_FAILURE);
  }
  if (fasta_outfile && phylip_outfile) {
    if (!(fasta_out = fopen(fasta_outfile, "w"))) {
      fprintf(stderr,"ERROR: bad output file name:\t%s\n", fasta_outfile);
      exit(EXIT_FAILURE);
    }
    if (!(phylip_out = fopen(phylip_outfile, "w"))) {
      fprintf(stderr,"ERROR: bad output file name:\t%s\n", phylip_outfile);
      exit(EXIT_FAILURE);
    }
  }
  else {
    fprintf(stderr,"ERROR: bad output files\n");
    exit(EXIT_FAILURE);
  }
  
  // Allocate the input sequence storage
  input_sequence = (char *)malloc(MAX_SEQ_LEN + 1);
  
  // Read read the sequences from the sequences file,
  // ignoring their names. Count the sequences with 'm'.
  while (readLine(in, buffer, BUFFER_SIZE) >= 0) {
    if (buffer[0] == '>') {
      
      // If we were reading a sequence (flag == 1), now we are done,
      // So convert the sequence
      if (flag) convert(input_sequence, fasta_out, phylip_out);
      
      // Otherwise we are about to print the name of the first file,
      // so print a blank line on line 1 of the phylip file to
      // save space for the sequence count and length.
      else fprintf(phylip_out, "%10s\n"," ");
    
      // Starting a new sequence -> clear the 'input_sequence'
      memset(input_sequence, 0, MAX_SEQ_LEN + 1);
      
      // Write the name of the sequence to the FASTA output file
      fprintf(fasta_out, "%s\n", buffer);
      
      // Write the name of the sequence to the Phylip output file
      // (the + 1 is so that the > doesnt get written.
      fprintf(phylip_out, "%-10s", buffer+1);
      
      // Increment the sequence count
      sequence_count++;

      // set flag indicating that we are now reading a sequence.
      flag = 1;
    }
    else if (flag) strcat(input_sequence, buffer);
  }
  if (flag) {
    sequence_length = (int)strlen(input_sequence);
    convert(input_sequence, fasta_out, phylip_out);
  }
  // Close the input file.
  fclose(in);
  
  // Move back to the beginning of the Phylip output file
  // in order to write the number of sequences and their length.
  rewind(phylip_out);
  fprintf(phylip_out, "%d %d", sequence_count, sequence_length);
  
  // free dynamic mem and close the output files.
  fclose(fasta_out);
  fclose(phylip_out);
  free(input_sequence);
}

/*******************************************
 * THE MAIN METHOD. FIRST GETS THE STRUCTURE
 * FROM THE STRUCTURE FILE, THEN CONVERTS
 * THE SEQUENCES.
 *******************************************

    ARGV[1]:               sequence file name
    ARGV[2](optional):     FASTA output file name
    ARGV[3](optional):     Phylip output file name   */

int main(int argc, char **argv) {
  	
  if (argc < 2 || argc > 4) {
    fprintf(stderr,
	    "USAGE:\n\n\t20to4 <sequence_file> [fasta_output_file] [phylip_output_file]\n\n"
	    "Input must be in FASTA format\nDefault output files are outfile.faa and outfile.phy\n");
    exit(EXIT_FAILURE);
  }
  
  if (argc > 2)
    if (argc > 3) 
      convert_sequences(argv[1], argv[2], argv[3]);
    else
      convert_sequences(argv[1], argv[2], "outfile.phy");
  else
    convert_sequences(argv[1], "outfile.faa", "outfile.phy");
  
  return EXIT_SUCCESS;
}
