/**************************************
Copyright Elisabeth Tillier and Andrew Smith, Univesity Health Network, 2003

 * PROGRAM TO CONVERT AN RNA SEQUENCE,
 * ALONG WITH A SECONDARY STRUCTURE 
 * (SET OF BASE PAIRINGS), INTO A 20 LETTER CODE.
 **************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

// The maximum number of pairs in a secondary structure.
#define MAX_PAIR 300000           

// The size of the input and output buffers.
#define BUFFER_SIZE 200000       
#define MAX_SEQ_LEN 200000

// Number of symbols in the original
// sequence (including structure).
#define CODE_COUNT 25           

// Number of paired bases specified in the structure file.
int pair_count;                   

// Holds info from the structure file 
// (in the form of pairs of indices)
int **base_pair;                  

// The input buffer for sequences.
char buffer[BUFFER_SIZE];

// The buffer to hold the converted sequence
// before it is printed.
char converted[BUFFER_SIZE];

// Mask to indicate whether a site was previously
// processed as part of a pair.
int *mask = NULL;

/*

This is the code:

A  -> A 
C  -> R 
G  -> N
U  -> D
AA -> C
AC -> Q 
AG -> E
AU -> G
CA -> H 
CC -> I
CG -> L
CU -> K
GA -> M
GC -> F
GG -> P
GU -> S
UA -> T
UC -> W
UG -> Y
UU -> V
*/

// The domain of the function from RNA to
// 20 letter code.
char *domain[25] = {
  "-A", "-C", "-G", "-U", "--",
  "AA", "AC", "AG", "AU", "A-", 
  "CA", "CC", "CG", "CU", "C-",
  "GA", "GC", "GG", "GU", "G-",
  "UA", "UC", "UG", "UU", "U-"
};

// The range of the function from RNA to
// 20 letter code.
char *range = 

 /*********************************
 *                                *
 *   -----AAAAACCCCCGGGGGUUUUU    *
 *   ACGU-ACGU-ACGU-ACGU-ACGU-    */
    "ARNDXCQEGAHILKRMFPSNTWYVD";
/*   0123456789012345678901234    *
 *                                *
 *********************************/

// The mapping for single bases.
inline char convert_base(char c) {
  if (c == 'A') return 'A';
  else if (c == 'C') return 'R';
  else if (c == 'G') return 'N';
  else if (c == 'U') return 'D';
  else if (c == 'X' || c == '-') return 'X';
  else {
    fprintf(stderr, "ERROR: bad character passed to convert_base()\n");
    exit(-1);
  }
  return '?';
}

/**********************************************
 *  CONVERT ANY DNA (T's) TO RNA AND MAKE SURE ALL
 *  ARE UPPER CASE
 **********************************************/
void DNA_to_RNA_and_upper_case(char *s) {
  int i = strlen(s)-1;
  for (; i >= 0; i--) {
    s[i] = toupper(s[i]);
    if (s[i] == 'T') s[i] = 'U';
  }
}

/********************************
 *  GIVEN THE INDEX OF A BASE,
 *  RETURN THE INDEX OF ITS PAIR
 ********************************/
int get_base_pair_index(int x) {
  int i;
  for (i = 0; i < pair_count; i++) 
    if (base_pair[0][i] == x) return i;
  return -1;
}

/*********************************
 *  GIVEN A PAIR OF BASES, RETURN THE
 *  CODE FOR THAT PAIR
 *********************************/
char find_code(char x, char y) {
  int i;
  
  // If one of the bases is unknown, code as 'U'
  if (x == 'X' || y == 'X') return 'X';
  
  // Iterate through 'domain' to find the correct index.
  for (i = 0; i < CODE_COUNT && 
	 (domain[i][0] != x || domain[i][1] != y); i++);
  
  // If the index was found, return the image of
  // the index.
  if (i < CODE_COUNT) return range[i];
  
  // Otherwise, bad input.
  fprintf(stderr, "Bad input caught in find_code()\n");
  exit(-1);
  
  return '?';
}

/********************************************************
 * READ A LINE OF TEXT FROM A FILE INTO A BUFFER
 ********************************************************/
int readLine(FILE *fp, char *buff, int buff_size) {
  int i;
  for (i = 0; i < buff_size && 
	 (buff[i] = fgetc(fp)) != EOF && 
	 (buff[i] != '\n') && (buff[i] != '\r'); i++);
  if (buff[i] == EOF) return -1;
  buff[i] = '\0';
  return i;
}

/*************************************
 * READ THE PAIRED BASES FROM THE STRUCTURE FILE
 *************************************/
void get_structure(char *structure_file) {
  FILE *structure;
  char buffer[BUFFER_SIZE];
  int i;
  
  // Open the structure file.
  if (!(structure = fopen(structure_file, "r"))) {
    fprintf(stderr,"ERROR: bad structure file name:\t%s\n", structure_file);
    exit(-1);
  }
  
  // Allocate the lookup table for paired bases.
  base_pair = (int **)malloc(2 * sizeof(int *));
  base_pair[0] = (int *)malloc(MAX_PAIR * sizeof(int));
  base_pair[1] = (int *)malloc(MAX_PAIR * sizeof(int));
  
  // Remove the header from the structure file
  for (i = 0; i < 4; i++) {
    readLine(structure, buffer, BUFFER_SIZE);
  }
  
  // Extract the base pairs from the input line.
  for (pair_count = 0; 
       fscanf(structure, "%d %*s %d\n", 
	      &base_pair[0][pair_count], 
	      &base_pair[1][pair_count]) > 0; 
       pair_count += (base_pair[1][pair_count] != 0));
  
  // Close the structure file
  fclose(structure);
}

/***********************************
 * CONVERT THE SEQUENCES INTO THE 20
 * SYMBOL CODE
 ***********************************/
void convert(char *original, FILE *fasta_out, FILE *phylip_out) {
  int length, pair_index, i = 0, j = 0;
  char x, y;
  
  // Convert to upper case, and get rid of the 'U's
  DNA_to_RNA_and_upper_case(original);
  
  // Get the length of the original string.
  length = strlen(original);
  
  // Allocate and zero the mask.
  if (!mask) mask = (int *)malloc(length * sizeof(int));
  memset(mask, 0, length * sizeof(int));
  //for (i = 0; i < length; i++) mask[i] = 0;
  
  // Convert each position...
  for (i = 0; i < length; i++) {
    // As long as the mask doesnt indicate that it
    // was converted previously as part of a pair.
    if (!mask[i]) {
      // The '+1' below is because the structure files index
      // sites in the sequence starting with '1', not '0'.
      if ((pair_index = get_base_pair_index(i+1)) != -1) {
	x = original[i];
	y = original[base_pair[1][pair_index]-1];
	converted[j] = find_code(x,y);
	mask[i] = 1;
	mask[base_pair[1][pair_index]-1] = 1;
      }
      else converted[j] = convert_base(original[i]);
      j++;
    }
  }
  converted[j] = '\0';
  fprintf(fasta_out, "%s\n", converted);
  fprintf(phylip_out, "%s\n", converted);
}

/***********************************
 * READ IN THE RNA SEQUENCES FROM A
 * FASTA FORMAT FILE.
 ***********************************/
void convert_sequences(char *sequence_file, char *fasta_outfile, char *phylip_outfile) {
  FILE *in, *fasta_out, *phylip_out; 
  int sequence_count = 0, sequence_length = 0, flag = 0;
  char *input_sequence;

  // Open the input and output files.
  if (!(in = fopen(sequence_file, "r"))) {
    fprintf(stderr,"ERROR: bad sequence file name:\t%s\n", sequence_file);
    exit(-1);
  }
  if (fasta_outfile && phylip_outfile) {
    if (!(fasta_out = fopen(fasta_outfile, "w"))) {
      fprintf(stderr,"ERROR: bad output file name:\t%s\n", fasta_outfile);
      exit(-1);
    }
    if (!(phylip_out = fopen(phylip_outfile, "w"))) {
      fprintf(stderr,"ERROR: bad output file name:\t%s\n", phylip_outfile);
      exit(-1);
    }
  }
  else {
    fprintf(stderr,"ERROR: bad output files\n");
    exit(-1);
  }
  
  // Allocate the input sequence storage
  input_sequence = (char *)malloc(MAX_SEQ_LEN + 1);
  
  // Read read the sequences from the sequences file,
  // ignoring their names. Count the sequences with 'm'.
  while (readLine(in, buffer, BUFFER_SIZE) >= 0) {
    if (buffer[0] == '>') {
      
      // If we were reading a sequence (flag == 1), now we are done,
      // So convert the sequence
      if (flag) convert(input_sequence, fasta_out, phylip_out);
      
      // Otherwise we are about to print the name of the first file,
      // so print a blank line on line 1 of the phylip file to
      // save space for the sequence count and length.
      else fprintf(phylip_out, "%10s\n"," ");
    
      // Starting a new sequence -> clear the 'input_sequence'
      memset(input_sequence, 0, MAX_SEQ_LEN + 1);
      
      // Write the name of the sequence to the FASTA output file
      fprintf(fasta_out, "%s\n", buffer);
      
      // Write the name of the sequence to the Phylip output file
      // (the + 1 is so that the > doesnt get written.
      fprintf(phylip_out, "%-10s", buffer+1);
      
      // Increment the sequence count
      sequence_count++;

      // set flag indicating that we are now reading a sequence.
      flag = 1;
    }
    else if (flag) strcat(input_sequence, buffer);
  }
  if (flag) {
    sequence_length = strlen(input_sequence);
    convert(input_sequence, fasta_out, phylip_out);
  }
  // Close the input file.
  fclose(in);
  
  // Move back to the beginning of the Phylip output file
  // in order to write the number of sequences and their length.
  rewind(phylip_out);
  fprintf(phylip_out, "%d %d", sequence_count, sequence_length);
  
  // and the output files.
  fclose(fasta_out);
  fclose(phylip_out);
}

/*******************************************
 * THE MAIN METHOD. FIRST GETS THE STRUCTURE
 * FROM THE STRUCTURE FILE, THEN CONVERTS
 * THE SEQUENCES.
 *******************************************/

/*

    ARGV[1]:               structure file name
    ARGV[2]:               sequence file name
    ARGV[3](optional):     FASTA output file name
    ARGV[4](optional):     Phylip output file name

                                                   */
int main(int argc, char **argv) {
  
  if (argc < 3 || argc > 5) {
    fprintf(stderr,
	    "USAGE:\n\n\t4to20 <structure_file> <sequence_file> [fasta_output_file] [phylip_output_file]\n\n"
	    "Input must be in FASTA format\nDefault output files are outfile.faa and outfile.phy\n");
    exit(-1);
  }
  get_structure(argv[1]);
  
  if (argv[3])
    if (argv[4]) 
      convert_sequences(argv[2], argv[3], argv[4]);
    else
      convert_sequences(argv[2], argv[3], "outfile.phy");
  else
    convert_sequences(argv[2], "outfile.faa", "outfile.phy");
  
  return 0;
}
